(* generated by dune *)

(** @canonical Dune_site.Dune_site_data *)
module Dune_site_data = Dune_site__Dune_site_data

(** @canonical Dune_site.Helpers *)
module Helpers = Dune_site__Helpers

module Dune_site__ = struct end
[@@deprecated "this module is shadowed"]
