(* generated by dune *)

(** @canonical Dune_rpc_private.Client *)
module Client = Dune_rpc_private__Client

(** @canonical Dune_rpc_private.Conv *)
module Conv = Dune_rpc_private__Conv

(** @canonical Dune_rpc_private.Dbus_address *)
module Dbus_address = Dune_rpc_private__Dbus_address

(** @canonical Dune_rpc_private.Diagnostics_v1 *)
module Diagnostics_v1 = Dune_rpc_private__Diagnostics_v1

(** @canonical Dune_rpc_private.Exported_types *)
module Exported_types = Dune_rpc_private__Exported_types

(** @canonical Dune_rpc_private.Fiber_intf *)
module Fiber_intf = Dune_rpc_private__Fiber_intf

(** @canonical Dune_rpc_private.Import *)
module Import = Dune_rpc_private__Import

(** @canonical Dune_rpc_private.Menu *)
module Menu = Dune_rpc_private__Menu

(** @canonical Dune_rpc_private.Procedures *)
module Procedures = Dune_rpc_private__Procedures

(** @canonical Dune_rpc_private.Public *)
module Public = Dune_rpc_private__Public

(** @canonical Dune_rpc_private.Registry *)
module Registry = Dune_rpc_private__Registry

(** @canonical Dune_rpc_private.Server_notifications *)
module Server_notifications = Dune_rpc_private__Server_notifications

(** @canonical Dune_rpc_private.Sub *)
module Sub = Dune_rpc_private__Sub

(** @canonical Dune_rpc_private.Types *)
module Types = Dune_rpc_private__Types

(** @canonical Dune_rpc_private.Version_error *)
module Version_error = Dune_rpc_private__Version_error

(** @canonical Dune_rpc_private.Versioned *)
module Versioned = Dune_rpc_private__Versioned

(** @canonical Dune_rpc_private.Versioned_intf *)
module Versioned_intf = Dune_rpc_private__Versioned_intf

(** @canonical Dune_rpc_private.Where *)
module Where = Dune_rpc_private__Where

module Dune_rpc_private__ = struct end
[@@deprecated "this module is shadowed"]
