(* generated by dune *)

(** @canonical Dune_site_plugins.Dune_site_plugins_data *)
module Dune_site_plugins_data = Dune_site_plugins__Dune_site_plugins_data

(** @canonical Dune_site_plugins.Meta_parser *)
module Meta_parser = Dune_site_plugins__Meta_parser

(** @canonical Dune_site_plugins.Plugins *)
module Plugins = Dune_site_plugins__Plugins

module Dune_site_plugins__ = struct end
[@@deprecated "this module is shadowed"]
