(* generated by dune *)

(** @canonical Dune_meta_parser.Meta_lexer *)
module Meta_lexer = Dune_meta_parser__Meta_lexer

(** @canonical Dune_meta_parser.Meta_parser *)
module Meta_parser = Dune_meta_parser__Meta_parser

module Dune_meta_parser__ = struct end
[@@deprecated "this module is shadowed"]
